@@ -1,8 +1,8 @@
-#define VERSION "1.54"
+#define VERSION "1.55"
 
 /*
  *  units, a program for units conversion
- *  Copyright (C) 1996, 1997 Free Software Foundation, Inc
+ *  Copyright (C) 1996, 1997, 1999 Free Software Foundation, Inc
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -452,7 +452,7 @@ productoverflow()
 
 /* Add a unit string onto the array of strings representing a product
    of units.  The new string will be added at the first NULLUNIT or at
-   the end (which is marked by a NULL pointer. */
+   the end (which is marked by a NULL pointer). */
 
 int
 addsubunit(char *product[], char *toadd)
@@ -460,7 +460,7 @@ addsubunit(char *product[], char *toadd)
    char **ptr;
 
    for (ptr = product; *ptr && *ptr != NULLUNIT; ptr++);
-   if (ptr >= product + MAXSUBUNITS) {
+   if (ptr - product >= MAXSUBUNITS - 1) {
       productoverflow();
       return 1;
    }
@@ -481,7 +481,7 @@ addsubunitlist(char *product[], char *toadd[])
    for(src = toadd; *src; src++){
      if (*src == NULLUNIT) continue;
      for(; *dest && *dest != NULLUNIT; dest++);
-     if (dest >= product + MAXSUBUNITS) {
+     if (dest - product >= MAXSUBUNITS - 1) {
        productoverflow();
        return 1;
      }
@@ -627,6 +627,52 @@ cancelunit(struct unittype *theunit)
    }
 }
 
+#if 0
+
+#include<math.h>
+
+/* 
+   The unitroot function takes the nth root of an input unit which has
+   been completely reduced.  Returns 1 if the unit is not a power of n. 
+   Input data can contain NULLUNITs.  
+*/
+
+int 
+subunitroot(int n,char *in[], char *out[])
+{
+  char **ptr,**result,**current;
+  int count;
+
+  result = out;
+  for(current = in;*current && *current==NULLUNIT;current++);
+  count = 0;
+  for(ptr=in;*ptr;ptr++){
+    if (*ptr==NULLUNIT) continue;
+    if (count==n) {
+      *(out++)=dupstr(*current);
+      current = ptr;
+      count=0;
+    }
+    if (!strcmp(*current,*ptr)) count++;
+    else return 1;
+  }
+  if (count==n) *(out++)=dupstr(*current);
+  else if (count!=0) return 1;
+  *out = 0;
+  return 0;
+}
+
+int 
+unitroot(int n,struct unittype *inunit, struct unittype *outunit)
+{
+   if ((n & 1==0) && (inunit->factor<0)) return 1;
+   outunit->factor = pow(inunit->factor,1.0/(double)n);
+   return subunitroot(n, inunit->numerator, outunit->numerator) &&
+           subunitroot(n, inunit->denominator, outunit->denominator);
+}
+
+#endif
+
 
 /*
    Looks up the definition for the specified unit including prefix processing
@@ -803,7 +849,7 @@ addunit(struct unittype *theunit, char *toadd, int flip)
    char *item;
    char *slash;
    int doingtop;
-
+   int handlesum(struct unittype *theunit, char *toadd, int flip);
 
    for(scratch=toadd;*scratch;scratch++)
      if (scratch>toadd && *scratch=='+' && !strchr("eE",*(scratch-1))) 
@@ -889,68 +935,6 @@ addunit(struct unittype *theunit, char *toadd, int flip)
 }
 
 
-/* 
-   Adds a unit expression which contains '+' characters to 
-   the unit structure by completely reducing each subexpression.
-*/
-
-int
-handlesum(struct unittype *theunit, char *toadd, int flip)
-{
-  struct unittype firstunit, nextunit;
-  char *first, *text, *rest;
-  int retval;
-  initializeunit(&firstunit);
-  first=dupstr(toadd);
-  rest=strchr(first,'+');   
-  if (rest) *rest++=0;      
-  if (isblankstr(first)){
-    free(first);
-    printf("Encountered '+' without operand.\n");
-    return 1;
-  }
-  if (addunit(&firstunit,first,flip) || completereduce(&firstunit)) {
-    free(first);
-    return 1;
-  }
-  while(rest){
-    text=rest;
-    rest=strchr(text,'+');
-    if (rest) *rest++=0;
-    if (isblankstr(text)){
-      printf("Encountered '+' without operand.\n");
-      return 1;
-    }
-    initializeunit(&nextunit);
-    if (addunit(&nextunit,text,flip) || completereduce(&nextunit)) {
-      free(first);
-      return 1;
-    }
-    if (compareunits(&firstunit,&nextunit)){
-      printf("Illegal sum of non-conformable units:\n");
-      first = removepadding(first);
-      text = removepadding(text);
-      printf("\t%s reduces to ",first);
-      showunit(&firstunit);
-      printf("\t%s reduces to ",text);
-      showunit(&nextunit);
-      free(first);
-      return 1;
-    }
-    firstunit.factor  += nextunit.factor;
-    freeunit(&nextunit);
-  }
-
-  retval = addsubunitlist(theunit->numerator,firstunit.numerator) ||
-    addsubunitlist(theunit->denominator,firstunit.denominator);
-  theunit->factor *= firstunit.factor;
-
-  freeunit(&firstunit);
-  free(first);
-  return retval;
-}
-
-
 /*
    reduces a product of symbolic units to primitive units.
    The three low bits are used to return flags:
@@ -1083,6 +1067,68 @@ completereduce(struct unittype *unit)
 }
 
 
+/* 
+   Adds a unit expression which contains '+' characters to 
+   the unit structure by completely reducing each subexpression.
+*/
+
+int
+handlesum(struct unittype *theunit, char *toadd, int flip)
+{
+  struct unittype firstunit, nextunit;
+  char *first, *text, *rest;
+  int retval;
+  initializeunit(&firstunit);
+  first=dupstr(toadd);
+  rest=strchr(first,'+');   
+  if (rest) *rest++=0;      
+  if (isblankstr(first)){
+    free(first);
+    printf("Encountered '+' without operand.\n");
+    return 1;
+  }
+  if (addunit(&firstunit,first,flip) || completereduce(&firstunit)) {
+    free(first);
+    return 1;
+  }
+  while(rest){
+    text=rest;
+    rest=strchr(text,'+');
+    if (rest) *rest++=0;
+    if (isblankstr(text)){
+      printf("Encountered '+' without operand.\n");
+      return 1;
+    }
+    initializeunit(&nextunit);
+    if (addunit(&nextunit,text,flip) || completereduce(&nextunit)) {
+      free(first);
+      return 1;
+    }
+    if (compareunits(&firstunit,&nextunit)){
+      printf("Illegal sum of non-conformable units:\n");
+      first = removepadding(first);
+      text = removepadding(text);
+      printf("\t%s reduces to ",first);
+      showunit(&firstunit);
+      printf("\t%s reduces to ",text);
+      showunit(&nextunit);
+      free(first);
+      return 1;
+    }
+    firstunit.factor  += nextunit.factor;
+    freeunit(&nextunit);
+  }
+
+  retval = addsubunitlist(theunit->numerator,firstunit.numerator) ||
+    addsubunitlist(theunit->denominator,firstunit.denominator);
+  theunit->factor *= firstunit.factor;
+
+  freeunit(&firstunit);
+  free(first);
+  return retval;
+}
+
+
 /* 
    If the given character string has only one unit name in it, then print out
    the rule for that unit.  In any case, print out the reduced form for
@@ -1215,6 +1261,33 @@ checkunits()
     }
 }
 
+
+#if 0
+
+/* 
+   Cycle through all units and print ones which are conformable with
+   the argument.
+*/
+
+void 
+tryallunits(struct unittype *have)
+{
+  struct unittype want;
+  struct unitlist *uptr;
+  int i;
+
+  for(i=0;i<HASHSIZE;i++)
+    for (uptr = utab[i]; uptr; uptr = uptr->next){
+      initializeunit(&want);
+      addunit(&want,uptr->name,0);
+      completereduce(&want);
+      if (!compareunits(have,&want))
+         printf("%s\n",uptr->name);
+      freeunit(&want);
+    }
+}
+#endif
+
 /* print usage message */
 
 void 
@@ -1274,6 +1347,24 @@ getuser(char **buffer, int *bufsize, char *query)
   }
 }
 
+/* Note:  What should this do?  
+
+   Complete either to the end of a prefix or complete to the 
+   end of a unit.  Or if there is a full prefix plus part of a unit, 
+   and if that prefix is longer than one character
+   then complete that compound.  Don't complete a prefix fragment into
+   prefix plus anything.  
+
+   Seems to be working right now...but one problem.  Completing a blank
+   entry always gets the cent sign.  Why is that?  My code seems to be
+   working ok....
+
+   One thing still needs to be done:  it won't complete a fragment into
+   a whole prefix.  It should be possible to type "myr<tab>" and
+   get "myria" for example.  
+*/
+   
+
 char *
 completeunits(char *text, int state)
 {
@@ -1282,7 +1373,7 @@ completeunits(char *text, int state)
   static int uhash;
   char *output,*thistry;
   
-  if (!state){
+  if (!state){     /* state = 0 means this is the first call, so initialize */
      uhash = 0;
      curprefix=0;
      curunit=utab[uhash];
@@ -1303,6 +1394,8 @@ completeunits(char *text, int state)
                         + (curprefix ? curprefix->len:0),"(completeunits)");
        strcpy(output,curprefix?curprefix->name:"");
        strcat(output,curunit->name);
+/*printf("     output='%s'\n",output);*/
+
     }
     curunit = curunit->next;
     while (!curunit  && uhash<HASHSIZE-1){
@@ -1311,8 +1404,10 @@ completeunits(char *text, int state)
     }
     if (!curunit && !curprefix){
       if (curprefix = plookup(text)){
-        uhash = 0;
-        curunit = utab[uhash];
+        if (strlen(curprefix->name)>1 && strlen(curprefix->name)<strlen(text)){
+          uhash = 0;
+          curunit = utab[uhash];
+        } else curprefix=0;
       }
     }
   }    
@@ -1436,9 +1531,8 @@ main(int argc, char **argv)
    int interactive;
 
 #ifdef READLINE
-/*   rl_bind_key ('\t', rl_insert);   Turn off filename completion */
-
    rl_completion_entry_function = (Function *)completeunits;
+   rl_basic_word_break_characters = " \t+-*/";
 #endif
 
    interactive = processargs(argc, argv, &havestr, &wantstr);
@@ -1458,6 +1552,12 @@ main(int argc, char **argv)
          showdefinition(havestr,&have);
          exit(0);
       }
+#if 0 
+      if (!strcmp(wantstr,"?")){    /* experimental code */
+         tryallunits(&have);
+         exit(0);
+      }
+#endif
       initializeunit(&want);
       if (addunit(&want, wantstr, 0) || completereduce(&want))
 	 exit(1);
